class Logger:
    def __init__(self, name, trigger=None) -> None:
        self.name = name
        self.trigger = trigger

    def print(self, *x, **kwargs):
        if self.trigger is not None:
            self.trigger({
                'type': 'print', 
                'name': self.name, 
                'args': x, 
                'kwargs': kwargs
            })
            return
        print('[{}]: '.format(self.name), end=' ')
        print(*x, **kwargs)

    def imshow(self, img):
        if self.trigger is not None:
            self.trigger({
                'type': 'imshow', 
                'name': self.name, 
                'args': img
            })
            return
        print(img)

    @classmethod
    def get_logger(cls, name):
        logger = cls(name, trigger=Logger.global_trigger)
        cls.loggers.setdefault(id(logger), logger)
        return logger

Logger.loggers = {}
Logger.global_trigger = None


from dlframe.CalculationNode import CalculationNode

class ExecutionNode:
    def __init__(self, node: CalculationNode) -> None:
        self.node = node
        self.in_degree = len(node.last_nodes)

        self.end_execution = False
        self.result = None

    def execute(self, config, node_dict):
        assert not self.end_execution, 'I do not know why this can happen'
        
        if self.node.is_root_node: # root nodes
            try:
                sub_config = config[self.node.node_manager.element_nodes[id(self.node)]]
                self.result = self.node.element_dict[sub_config]
            except:
                assert len(self.node.element_dict) == 1, f'no config for {self.node.node_manager.element_nodes[id(self.node)]}'
                self.result = next(iter(self.node.element_dict.values()))
            
        elif self.node.is_function: # function nodes
            args = [node_dict[id(sub_node)].result for sub_node in self.node.function_args]
            kwargs = {kw: node_dict[id(self.node.function_kwargs[kw])].result for kw in self.node.function_kwargs.keys()}
            if len(args) + len(kwargs) == len(self.node.last_nodes): # pure function node or init function node
                current_function = None
                try:
                    sub_config = config[self.node.node_manager.element_nodes[id(self.node)]]
                    current_function = self.node.element_dict[sub_config]
                except:
                    assert len(self.node.element_dict) == 1, f'no config for {self.node.node_manager.element_nodes[id(self.node)]}'
                    current_function = next(iter(self.node.element_dict.values()))
                self.result = current_function(*args, **kwargs)
            else: # class method node
                self.result = node_dict[id(self.node.last_nodes[0])].result(*args, **kwargs)
        else: # attr nodes
            self.result = getattr(node_dict[id(self.node.last_nodes[0])].result, self.node.attr_name)

        self.end_execution = True


class CalculationNode:
    def __init__(self, attr_name: str, node_manager, last_node=None, element_dict: dict=None, is_root_node=False) -> None:
        self.is_root_node = is_root_node
        
        self.attr_name = attr_name
        self.last_nodes = [last_node] if last_node is not None else []
        self.next_nodes = []
        self.node_manager = node_manager
        self.element_dict = element_dict
        
        self.is_function = False
        self.is_init_function = False
        self.function_args = []
        self.function_kwargs = {}

        self.node_manager._register_node(self)
    
    def __call__(self, *args, **kwargs):
        all_args = []
        for arg in args:
            if type(arg) is not CalculationNode:
                arg = CalculationNode(str(arg), self.node_manager, None, {str(arg): arg}, is_root_node=True)
            all_args.append(arg)

        all_kwargs = {}
        for kw in kwargs.keys():
            arg = kwargs[kw]
            if type(arg) is not CalculationNode:
                arg = CalculationNode(str(arg), self.node_manager, None, {str(arg): arg}, is_root_node=True)
            all_kwargs.setdefault(kw, arg)

        node = CalculationNode('function_call', self.node_manager, self)
        self.next_nodes.append(node)

        node.is_function = True
        node.function_args = all_args
        node.function_kwargs = all_kwargs

        if self.element_dict is not None and len(self.element_dict) > 0 and type(next(iter(self.element_dict.values()))) == type:
            for t in self.element_dict.values():
                assert type(t) == type, 'elements should have same type'
            node.is_init_function = True

        for arg in all_args:
            arg.next_nodes.append(node)
            node.last_nodes.append(arg)
        for kw in all_kwargs.keys(): 
            arg = kwargs[kw]
            arg.next_nodes.append(node)
            node.last_nodes.append(arg)
        
        return node
    
    def __getattr__(self, __name: str):
        if __name == '__iter__':
            return None

        node = CalculationNode(__name, self.node_manager, self)
        self.next_nodes.append(node)
        return node
    
    def __getitem__(self, index):
        node = CalculationNode('__getitem__', self.node_manager, self)
        self.next_nodes.append(node)
        
        return node(index)
    
    def __gt__(self, other):
        self.next_nodes.append(other)
        other.last_nodes.append(self)
        return self

    def __lt__(self, other):
        self.last_nodes.append(other)
        other.next_nodes.append(self)
        return self


import queue
from dlframe.CalculationNode import CalculationNode
from dlframe.ExecutionNode import ExecutionNode

class CalculationNodeManager:
    def __init__(self, parallel=False) -> None:
        self.nodes = {}
        self.element_nodes = {}
        self.class_type_name2node = {}

        self.latest_node = None
        self.parallel = parallel

    def _register_node(self, node: CalculationNode):
        self.nodes.setdefault(id(node), node)
        if not self.parallel:
            if self.latest_node is not None:
                self.latest_node > node
            self.latest_node = node
        return self

    def register_element(self, name: str, element_dict: dict=None, display: bool=True):
        if element_dict is None:
            element_dict = {}
        node = CalculationNode('__Element__' + str(element_dict), self, None, element_dict, is_root_node=True)
        self.class_type_name2node.setdefault(name, node)
        if not display:
            assert len(element_dict) == 1, 'you can only register one element if not display'
        else:
            self.element_nodes.setdefault(id(node), name)
        return node

    def inspect(self):
        name_dict = {node_name: list(self.nodes[node_id].element_dict.keys()) for node_id, node_name in self.element_nodes.items()}
        return name_dict

    def execute(self, config: dict, max_size=0):
        node_dict = {name: ExecutionNode(node) for name, node in self.nodes.items()}
        node_queue = queue.Queue(maxsize=max_size)
        for name, node in node_dict.items():
            if node.in_degree == 0:
                node_queue.put(name)

        while not node_queue.empty():
            current_node_name = node_queue.get()
            current_node = node_dict[current_node_name]

            current_node.execute(config, node_dict)
            
            for next_node in current_node.node.next_nodes:
                next_node_name = id(next_node)
                node_dict[next_node_name].in_degree -= 1
                if node_dict[next_node_name].in_degree == 0:
                    node_queue.put(next_node_name)
        return self
    
    def register(self, class_type_name, option_name=None, element=None, display=None):
        if type(option_name) == dict:
            if display is None:
                display = True
            return self.register_element(class_type_name, option_name, display)
        if element is not None:
            if option_name is None:
                option_name = element.__name__
            node = None
            if class_type_name not in self.class_type_name2node.keys():
                element_dict = {option_name: element}
                if class_type_name is None:
                    class_type_name = '__Element__' + str(element_dict)
                elif display is None:
                    display = True
                node = CalculationNode(class_type_name, self, None, element_dict, is_root_node=True)
                self.class_type_name2node.setdefault(class_type_name, node)
                if display:
                    self.element_nodes.setdefault(id(node), class_type_name)
            else:
                node = self.class_type_name2node[class_type_name]
                node.element_dict.setdefault(option_name, element)
                if display is not None:
                    assert False, f'The display option is only supported when registering for the first time. Please set the display option when registering element "{next(iter(node.element_dict.keys()))}".'
            return node
        def _warp_function_(raw_element, class_type_name=class_type_name, option_name=option_name, display=display, self=self):
            self.register(class_type_name, option_name, raw_element, display)
            return raw_element
        return _warp_function_
    
    def __getitem__(self, key):
        if key in self.class_type_name2node.keys():
            return self.class_type_name2node[key]
        return self.register(key, {}, display=True)


import queue
import threading
import asyncio
import json
import websockets
import traceback

from dlframe.CalculationNodeManager import CalculationNodeManager
from dlframe.Logger import Logger

class SendSocket:
    def __init__(self, socket) -> None:
        self.sendBuffer = queue.Queue()
        self.socket = socket
        self.sendThread = threading.Thread(target=self.threadWorker, daemon=True)
        self.sendThread.start()

    def threadWorker(self):
        event_loop = asyncio.new_event_loop()
        asyncio.set_event_loop(event_loop)
        event_loop.run_until_complete(self.sendWorker())

    async def sendWorker(self):
        while True:
            content = self.sendBuffer.get()
            await self.socket.send(content)

    def send(self, content: str):
        self.sendBuffer.put(content)

class WebManager(CalculationNodeManager):
    def __init__(self, host='0.0.0.0', port=8765, parallel=False) -> None:
        super().__init__(parallel=parallel)
        self.host = host
        self.port = port

    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            raise exc_val
        self.start(self.host, self.port)

    def start(self, host=None, port=None) -> None:
        if host is None:
            host = self.host
        if port is None:
            port = self.port
        event_loop = asyncio.new_event_loop()
        asyncio.set_event_loop(event_loop)
        async def onRecv(socket, path):
            msgIdx = -1
            sendSocket = SendSocket(socket)
            async for message in socket:
                msgIdx += 1
                message = json.loads(message)
                # print(msgIdx, message)

                # key error
                if not all([key in message.keys() for key in ['type', 'params']]):
                    response = json.dumps({
                        'status': 500, 
                        'data': 'no key param'
                    })
                    await socket.send(response)
                
                # key correct
                else:
                    if message['type'] == 'overview':
                        response = json.dumps({
                            'status': 200, 
                            'type': 'overview', 
                            'data': self.inspect()
                        })
                        await socket.send(response)

                    elif message['type'] == 'run':
                        params = message['params']
                        conf = params

                        def image2base64(img):
                            import base64
                            from io import BytesIO
                            from PIL import Image

                            # 创建一个示例NumPy数组（图像）
                            image_np = img

                            # 将NumPy数组转换为PIL.Image对象
                            image_pil = Image.fromarray(image_np)

                            # 将PIL.Image对象保存为字节流
                            buffer = BytesIO()
                            image_pil.save(buffer, format='JPEG')
                            buffer.seek(0)

                            # 使用base64库将字节流编码为base64字符串
                            image_base64 = base64.b64encode(buffer.read()).decode('utf-8')

                            return image_base64

                        Logger.global_trigger = lambda x: sendSocket.send(json.dumps({
                            'status': 200, 
                            'type': x['type'], 
                            'data': {
                                'content': '[{}]: '.format(x['name']) + ' '.join([str(_) for _ in x['args']]) + getattr(x['kwargs'], 'end', '\n') if x['type'] == 'print' \
                                    else image2base64(x['args'])
                            }
                        }))
                        for logger in Logger.loggers.values():
                            if logger.trigger is None:
                                logger.trigger = Logger.global_trigger

                        try:
                            self.execute(conf)
                        except Exception as e:
                            response = json.dumps({
                                'status': 200, 
                                'type': 'print', 
                                'data': {
                                    'content': traceback.format_exc()
                                }
                            })
                            await socket.send(response)
                    
                    # unknown key
                    else:
                        response = json.dumps({
                            'status': 500, 
                            'data': 'unknown type'
                        })
                        await socket.send(response)

        print('The backend server is running on [{}:{}]...'.format(host, port))
        print('The frontend page is at: https://picpic2013.github.io/dlframe-front/')

        event_loop.run_until_complete(websockets.serve(onRecv, host, port))
        event_loop.run_forever()


from dlframe.CalculationNodeManager import CalculationNodeManager
from dlframe.WebManager import WebManager
from dlframe.Logger import Logger

from dlframe.TimeWarpper import PICTimer
from dlframe.RecursiveWarpper import make_recursive_func, make_multi_return_recursive_func


from __future__ import annotations
import time
from typing import Any, Callable, Union

class PICTimerWith:
    def __init__(self, name=None, showTimeWhenRunning=False, autoStart=False) -> None:
        '''
        @param name:                timer name ('untitled' in default)
        @param showTimeWhenRunning: whether enable print before summary
               Note that printing the results before the summary will 
               reduce the time measurement accuracy
        '''
        if name is None:
            name = 'untitled'
        self.__timer_name = name
        self.__showCount = 0
        self.__intervalTimes = []
        self.__startTime = None
        
        self.__results = []
        self.__showTimeWhenRunning = showTimeWhenRunning

        self.__subTimers = []
        
        if autoStart:
            self.startTimer()

    def __printToResults(self, *content, end='\n'):
        if self.__showTimeWhenRunning:
            print(*content, end=end)
        else:
            self.__results.append((content, end))
            # self.__results.append(''.join(content) + end)

    def showTime(self, name=None, forcePrint=False) -> PICTimerWith:
        '''
        log time
        @param name:       time point name ('stage_%d' in default)
        @param forcePrint: whether print immediately (SLOW)
        @returns:          the timer
        '''
        assert self.__startTime is not None, 'you should call ${timer_name}.startTimer() first!'

        if name is None:
            name = 'stage_%d' % (self.__showCount)
        
        endTime = time.time()
        self.__printToResults("\033[32m[ {}.{} ] excution time(from begin): \033[31m{:.4f} ms\033[0m".format(self.__timer_name, name, (endTime - self.__startTime) * 1000), end='')
        if self.__showCount > 0:
            self.__printToResults(", \033[32m[ {}.{} ] excution time(from last): \033[31m{:.4f} ms\033[0m".format(self.__timer_name, name, (endTime - self.__intervalTimes[-1]) * 1000))
        else:
            self.__printToResults()
        self.__intervalTimes.append(endTime)
        self.__showCount += 1

        if forcePrint:
            self.forceShowTime()

        return self
    
    def forceShowTime(self) -> PICTimerWith:
        '''
        show time immediately
        Note that this function will reduce the time measurement accuracy
        Do NOT use this function in ANY time sensitive situation!
        '''
        for c, e in self.__results:
            print(*c, end=e)
        self.__results.clear()
        return self

    def startTimer(self) -> PICTimerWith:
        '''
        start the timer
        '''
        assert self.__startTime is None, 'the timer is already started!'
        self.__startTime = time.time()
        self.__intervalTimes.append(self.__startTime)
        return self

    def summary(self) -> PICTimerWith:
        '''
        quit the timer and show results
        '''

        assert self.__startTime is not None, 'you should call ${timer_name}.startTimer() first!'

        for subTimer in self.__subTimers:
            subTimer.summary()

        self.showTime('total' if self.__showCount == 0 else 'last')
        if len(self.__intervalTimes) > 2:
            deltaTimes = [j - i for i, j in zip(self.__intervalTimes, self.__intervalTimes[1:])]
            self.__printToResults("\033[32m[ {}.{} ] excution time: \033[31m{:.4f} ms\033[0m".format(self.__timer_name, 'mean', sum(deltaTimes) * 1000 / len(deltaTimes)))
        self.__printToResults()

        if not self.__showTimeWhenRunning:
            self.forceShowTime()
        return self

    def __enter__(self) -> PICTimerWith:
        return self.startTimer()

    def __exit__(self, type, value, trace):
        self.summary()

    def getTimer(self, name=None, showTimeWhenRunning=False, autoStart=False) -> PICTimerWith:
        if name is None:
            name = 'untitled'
        subTimer = PICTimerWith(self.__timer_name + '.' + name, showTimeWhenRunning=showTimeWhenRunning, autoStart=autoStart)
        self.__subTimers.append(subTimer)
        return subTimer

def makePICTimerWrapper(func) -> Union[Callable, PICTimer]:
    return func

@makePICTimerWrapper
class PICTimer:
    def __init__(self, func: Callable) -> None:
        self.func = func

    def call(self, *args: Any, **kwargs: Any) -> Any:
        assert self.func is not None, 'function is None'

        with PICTimer.getTimer(self.func.__name__):
            res = self.func(*args, **kwargs)
        return res

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        return self.call(*args, **kwds)

    @staticmethod
    def getTimer(name=None, showTimeWhenRunning=False, autoStart=False) -> PICTimerWith:
        return PICTimerWith(name, showTimeWhenRunning=showTimeWhenRunning, autoStart=autoStart)


# examples
if __name__ == '__main__':
    # example 1
    # function timer
    print('='*50, 'example 1', '='*50)
    @PICTimer
    def f1():
        for _ in range(100000):
            b = _

    f1()

    # example 2
    # with code block
    print('='*50, 'example 2', '='*50)
    with PICTimer.getTimer('example2') as t:
        for _ in range(1, 100000):
            if _ % 20000 == 0:
                t.showTime()
            bb = _

    # example 3
    print('='*50, 'example 3', '='*50)
    timer = PICTimer.getTimer('example3') # -┬--> or `timer = PICTimer.getTimer('example3', autoStart=True)`
    timer.startTimer()                    # -┘
    for idx in range(3):
        for _ in range(1, 100000):
            if _ % 20000 == 0:
                timer.showTime("stage_" + str(_))
            bb = _
        timer.forceShowTime() # you can output results before summary
    timer.summary()

    # example 4
    # create sub-timer
    print('='*50, 'example 4', '='*50)
    timer = PICTimer.getTimer('example4', autoStart=True)
    for idx in range(3):
        subTimer = timer.getTimer('sub_' + str(idx), autoStart=True)
        for _ in range(1, 100000):
            if _ % 20000 == 0:
                subTimer.showTime("stage_" + str(_))
            bb = _
        timer.showTime()
    timer.summary()


from typing import Callable
from functools import reduce, wraps

def decohints(decorator: Callable) -> Callable:
    return decorator

def isTheSameType(last, now):
    '''
    @param last: (lastRes, lastType)
        lastRes: None | are previous elements same type
        lastType: type of the last element

    @param now:  current element
    @returns:  (lastRes, lastType)
    '''
    lastRes, lastType = last
    return (lastType is None or (lastRes and (lastType == type(now))), type(now))

@decohints
def make_recursive_func(func: Callable) -> Callable:
    """Convert a function into recursive style to handle nested dict/list/tuple variables

    Args:
        func: input function
        func Args: 
            vars:   element / list / tuple / dict
            args:   other arguments
            kwargs: other keyword arguments

    Returns:
        recursive style function
    """

    def tupleGeneration(*args, **kwargs):
        for arg in zip(*args):
            yield wrapper(*arg, **kwargs)

    @wraps(func)
    def wrapper(*args, **kwargs):
        eleEq, eleType = reduce(isTheSameType, args, (True, None))
        if eleEq:
            if eleType == list:
                return [wrapper(*arg, **kwargs) for arg in zip(*args)]
            elif eleType == tuple or eleType == type(tupleGeneration()):
                return tupleGeneration(*args, **kwargs)
            elif eleType == dict:
                resDict = {}
                for arg in args:
                    for k, v in arg.items():
                        resDict.setdefault(k, [])
                        resDict[k].append(v)
                return {k:wrapper(*v, **kwargs) for k, v in resDict.items()}
        return func(*args, **kwargs)
    return wrapper

@decohints
def make_multi_return_recursive_func(func: Callable) -> Callable:
    """Convert a function into recursive style to handle nested dict/list/tuple variables
    that returns multiple results
    WARNING! tuple is treated the same as list

    Args:
        func: input function
        func Args: 
            vars:   element / list / tuple / dict
            args:   other arguments
            kwargs: other keyword arguments

    Returns:
        recursive style function
    """

    def tupleGeneration(*args, **kwargs):
        for arg in zip(*args):
            yield wrapper(*arg, **kwargs)

    @wraps(func)
    def wrapper(*args, **kwargs):
        eleEq, eleType = reduce(isTheSameType, args, (True, None))
        if eleEq:
            if eleType == list or (eleType == tuple or eleType == type(tupleGeneration())):
                tempReturn = [wrapper(*arg, **kwargs) for arg in zip(*args)]
                returnResults = None
                for item in tempReturn:
                    if returnResults is None:
                        returnResults = [[_] for _ in item]
                    else:
                        assert len(returnResults) == len(item), 'the function must have the same number of the return value'
                        for oldResult, newResult in zip(returnResults, item):
                            oldResult.append(newResult)
                return returnResults
            elif eleType == dict:
                resDict = {}
                for arg in args:
                    for k, v in arg.items():
                        resDict.setdefault(k, [])
                        resDict[k].append(v)
                tempReturn = {k:wrapper(*v, **kwargs) for k, v in resDict.items()}
                returnResults = None
                for key, item in tempReturn.items():
                    if returnResults is None:
                        returnResults = [{key: value} for value in item]
                    else:
                        assert len(returnResults) == len(item), 'the function must have the same number of the return value'
                        for oldResult, newValue in zip(returnResults, item):
                            oldResult.setdefault(key, newValue)
                return returnResults

        return func(*args, **kwargs)
        
    return wrapper

if __name__ == '__main__':
    # dict args
    arg1 = {idx: 'arg1_k_'+str(idx) for idx in range(3)}
    arg2 = {idx: 'arg2_k_'+str(idx) for idx in range(3)}
    
    # list args
    arg1 = [_ for _ in range(10)]
    arg2 = [_ for _ in range(10)]

    # tuple / generator args
    arg1 = (_ for _ in range(10))
    arg2 = (_ for _ in range(10))

    # non-iterable args
    arg1 = 0
    arg2 = 1

    # mutiple args
    arg1 = {idx: ['arg1_k_'+str(idx)+'_l_'+str(_) for _ in range(2)] for idx in range(3)}
    arg2 = {idx: ['arg2_k_'+str(idx)+'_l_'+str(_) for _ in range(2)] for idx in range(3)}

    @make_recursive_func
    def f1(x, base):
        return base.format(x)

    @make_recursive_func
    def f2(x, y, base):
        return base.format(x, y)

    @make_recursive_func
    def f3(x, y, base1, base2):
        return base1.format(x), base2.format(y)

    @make_multi_return_recursive_func
    def f4(x, y, base1, base2):
        return base1.format(x), base2.format(y)

    res1 = f1(arg1, base="arg1: {}")
    res2 = f2(arg1, arg2, base="arg1: {}, arg2: {}")
    res3 = f3(arg1, arg2, base1="arg1: {}", base2=" arg2: {}")
    res4 = f4(arg1, arg2, base1="arg1: {}", base2=" arg2: {}")

    print(res1)
    print(res2)
    print(res3)
    print(res4)
    print()


from setuptools import setup
import pathlib
import os

here = pathlib.Path(__file__).parent.resolve()

with open(os.path.join(here, 'README.md'), 'r',encoding='UTF-8') as mdFile:
    long_description = mdFile.read()

setup(
    name='dynamic-links', 
    version='0.1.1', 
    packages=['dlframe'], 
    url='https://github.com/picpic2013/dlframe-back', 
    license='MIT', 
    author='PIC', 
    author_email='picpic2019@gmail.com', 
    description='a calculation framework', 
    long_description=long_description, 
    long_description_content_type="text/markdown",
    python_requires=">=3.8, <4", 
    install_requires=[
        'websockets>=10.3', 'numpy', 'Pillow'
    ]
)
